#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import numpy as np
import scipy.ndimage as ndi
import scipy.optimize as opt

import doas

from smartcarb import io
from smartcarb import misc
from smartcarb import plume


# kg/s -> Mt/y (Tg/y)
kgs_to_Mty = 1e3 * (365 * 24 * 60 * 60) / 1e12


def compute_emissions(nc):

    cats = [
       'CO2_point_snap1_major', # this was counted twice in Jan and Jul
       'CO2_point_snap1_minor',
       'CO2_point_snap2_minor',
       'CO2_point_snap3_minor',
       'CO2_point_snap4_minor',
       'CO2_point_snap5_minor',
       'CO2_point_snap34_minor',
       'CO2_point_snap8_minor',
       'CO2_point_snap9_minor',
       'CO2_area_construction',
       'CO2_line_aircraft',
       'CO2_point_aircraft',
       'CO2_area_aircraft',
       'CO2_area_forestry',
       'CO2_area_offroad',
       'CO2_line_shipping',
       'CO2_point_shipping',
       'CO2_line_train',
       'CO2_area_heating',
       'CO2_line_majorroads',
       'CO2_area_minorroads'
    ]

    total = 0.0
    for cat in cats:
        total += np.sum(nc.variables[cat][:] * 1235941.1)

    return total



def get_constant_emissions(time, use_constant=True):
    if not use_constant and time.year == 2015 and time.month in [1,7]:
        return 754.86
    else:
        return 534.30



def simple_leastsq(filename, tracer, thr, cloud_thr=0.01, error_std=0.7, seed='orbit',
        background_mode='model', use_constant_emissions=False, use_real_footprint=False
    ):
    """\
    filename is either a file with 2d COSMO field or Level-2 data

    y[i,j] = x * H[i,j] + b

    - y: XCO2 measurements in plume
    - H: XCO2 from Berlin tracer (e.g. XCO2_BC) normalized by emissions
    - b: background (e.g. mean of "background" tracers in COSMO)

    The plume is defined where 'plume_tracer' > 'plume_thr'.

    If `use_const_emissions` use XCO2_BV in observation vector y.
    """
    # tracer fields (synthetic observations)
    time = io.filename2time(filename)

    # define plume based on threshold
    p = plume.GroundTruth(filename, 'Berlin', 'CO2', 0.05, 0.3, 10)
    is_plume = p.plume & (p.signal > thr) & (p.clouds < cloud_thr)


    bv, bc = io.read_fields(filename, ['XCO2_BV', 'XCO2_BC'], time, correct_berlin=True)

    co2 = io.read_total_tracer(filename, 'CO2', time, correct_berlin=True,
            use_constant_emissions=use_constant_emissions)

    if seed == 'orbit':
        _, orbit, _, lon_eq = io.process_filename(filename)
        np.random.seed(orbit + (lon_eq // 90 * 6000))

    noise = io.read_tracer_noise(filename, tracer, True, error_std)



    if np.sum(is_plume) <= 1:
        return np.nan, np.nan, np.sum(is_plume)


    # background
    if use_constant_emissions:
        b = (co2 - bc)[is_plume]
    else:
        b = (co2 - bv)[is_plume]


    if background_mode == 'model':
        bg_std = 0.0
    elif background_mode == 'mean':
        bg_std = b.std()
        b = b.mean()
    else:
        raise ValueError('unknown mode for background estimation')

    # observation vector
    y = co2[is_plume]

    # footprint
    if use_real_footprint and not use_constant_emissions:
        H = bv[is_plume] / 534.30
    else:
        H = bc[is_plume] / 534.30

    # add noise
    y += noise[is_plume]

    # Se, xa and Sa
    Se = np.nan

    # forward model
    f = lambda x,_: H * x / kgs_to_Mty + b

    x0 = np.array([534.0]) * kgs_to_Mty
    x, res = doas.solver.gauss_newton(x0, y, f, Se=Se)

    return x[0], np.sqrt(float(res['S'])), np.sum(is_plume)






