#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import osr, ogr
import numpy as np
import scipy
import scipy.ndimage as ndi
import cartopy.crs as ccrs
import shapely.geometry

from smartcarb import config as cfg



class DataField:
    def __init__(self, x, y, c, origin, crs='wgs84', pollon=cfg.pollon,
                 pollat=cfg.pollat, epsg=cfg.EPSG):
        """
        data field: c = f(x,y)

        transform: 'wgs84', 'rot' or 'xy'
        """
        self.data = c

        # coordinate system parameters
        self._origin = origin
        self._epsg = epsg
        self._pollon = pollon
        self._pollat = pollat

        # transform x,y to WGS 84
        if crs == 'wgs84':
            self.lon, self.lat = x, y
        elif crs == 'rot':
            self.lon, self.lat = rotpole2wgs(x, y, self.pollon, self.pollat)
        elif crs == 'xy':
            self.lon, self.lat = xy2wgs(x, y, origin, epsg)
        else:
            raise ValueError('Unknown crs "%s".' % crs)

        self._update()


    def _update(self):
        """ Update x,y and and rlon, rlat."""
        self.x, self.y = xy2wgs(self.lon, self.lat, self._origin, self._epsg,
                                inverse=True)
        self.rlon, self.rlat = rotpole2wgs(self.lon, self.lat, self._pollon,
                                           self._pollat, inverse=True)


    def interp_xy(self, x,y):
        """
        Get data value at x,y coords using linear interpolation.
        """
        points = np.transpose([self.x.flatten(), self.y.flatten()])
        interp = scipy.interpolate.LinearNDInterpolator(points, self.data.flatten())

        return interp(x,y)


    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = value
        self._update()

    @property
    def epsg(self):
        return self._epsg

    @epsg.setter
    def epsg(self, value):
        self._epsg = value
        self._update()

    @property
    def pollon(self):
        return self._pollon

    @pollon.setter
    def pollon(self, value):
        self._pollon = value
        self._update()

    @property
    def pollat(self):
        return self._pollat

    @pollat.setter
    def pollat(self, value):
        self._pollat = value
        self._update()




def get_source_origin(name):
    if name == 'Berlin':
        lon_o, lat_o = cfg.CITIES[name]
    else:
        lon_o, lat_o = cfg.POINT_SOURCES[name]

    return lon_o, lat_o



def get_circle(lon_o, lat_o, radius=5.0):
    """
    Return circle around origin with radius (in km).
    """
    radius = radius / 100.0 # km -> degree (roughly)
    p = shapely.geometry.Point(lon_o, lat_o)

    if radius > 0:
        p = p.buffer(radius)

    return p




def compute_centroid(x, y, w):
    """\
    Compute center of plume.
    """
    x0 = np.sum(w * x) / np.sum(w)
    y0 = np.sum(w * y) / np.sum(w)

    return x0, y0



def anglediff(alpha, beta, is_degree=True):
    if is_degree:
        alpha = np.deg2rad(alpha)
        beta = np.deg2rad(beta)

    delta = np.arctan2(
        np.sin(alpha-beta),
        np.cos(alpha-beta)
    )
    return np.rad2deg(delta) if is_degree else delta


def nearest_index(lon_o, lat_o, lon, lat):
    index = np.argmin(np.abs(lon - lon_o) + np.abs(lat - lat_o))
    i,j = np.unravel_index(index, lon.shape)

    return i,j


def wind_direction(u, v):
    return (270.0 - np.rad2deg(np.arctan2(v,u))) % 360


def iter_times(start, end, step=timedelta(hours=1), exclude_end=False):
    current = start

    if exclude_end:
        while current < end:
            yield current
            current = current + step
    else:
        while current <= end:
            yield current
            current = current + step



def create_rotpole_reference(pollon, pollat):
    """\
    Return spatial reference for rotated pole coordinates.

    pollon: longitude of north pole
    pollat: latitude of north pole
    """
    ref = osr.SpatialReference()
    ref.ImportFromEPSG(4326)     # WSG-84

    if (pollon is not None) and (pollat is not None):
        o_lon_p = 0.0
        o_lat_p = pollat
        lon_0 = 180 + pollon

        string = (
            'PROJCS["rotated_pole projection",EXTENSION["PROJ4","+proj=ob_tran +o_proj=longlat'
            ' +o_lon_p=%.2f +o_lat_p=%.2f +lon_0=%.2f +a=1 +to_meter=0.0174532925199"]]'
        ) % (o_lon_p, o_lat_p, lon_0)

        ref.ImportFromWkt(string)    # rotpole transformation

    return ref




def xy2wgs(x, y, origin, epsg=cfg.EPSG, inverse=False):
    """
    xy (kilometers from origin) to WGS84.
    """
    lon_o, lat_o = get_source_origin(origin)
    x_o, y_o = wgs2epsg(lon_o, lat_o, epsg)

    if inverse:
        x,y = wgs2epsg(x,y, epsg, inverse=False)
        return (x - x_o) / 1e3, (y - y_o) / 1e3

    else:
        x = 1e3 * x + x_o # in km
        y = 1e3 * y + y_o # in km

        return wgs2epsg(x,y, epsg, inverse=True)


def wgs2epsg(lon, lat, epsg=cfg.EPSG, inverse=False):
    """
    Transforms lon/lat to EPSG.
    """
    if inverse:
        out = ccrs.PlateCarree()
        in_ = ccrs.epsg(epsg)
    else:
        out = ccrs.epsg(epsg)
        in_ = ccrs.PlateCarree()

    if np.ndim(lon) == 0:
        res = out.transform_point(lon, lat, in_)
        return res[0], res[1]
    else:
        res = out.transform_points(in_, lon, lat)
        return res[...,0], res[...,1]


def rotpole2wgs(rlon, rlat, pollon=cfg.pollon, pollat=cfg.pollat,
                inverse=False):
    """
    Transform rotated pole to WGS84.
    """
    c_in = ccrs.RotatedPole(pollon, pollat)
    c_out  = ccrs.PlateCarree()

    if inverse:
        c_in, c_out = c_out, c_in

    if np.ndim(rlon) == 0:
        res = c_out.transform_point(rlon, rlat, c_in)
        return res[0], res[1]
    else:
        res = c_out.transform_points(c_in, rlon, rlat)
        return res[...,0], res[...,1]



def spatial_binning(a, n=2):
    if np.ndim(a) == 2:
        kernel = np.ones((n,n)) / n**2
        return ndi.convolve(a, kernel)[::n,::n]
    else:
        kernel = np.ones((1,n,n)) / n**2
        return ndi.convolve(a, kernel)[:,::n,::n]


